from typing import Any, Dict

DJANGO_PYOIDC: Dict[str, Dict[str, Any]] = {}

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "django.contrib.sessions",
    "django_pyoidc",
]



from typing import Any, Dict, List, Optional, TypedDict

from django.urls import path

ProviderConfig = TypedDict(
    "ProviderConfig",
    {
        # most important ----
        "client_id": Optional[str],
        "client_secret": Optional[str],
        "oidc_cache_provider_metadata": Optional[str],
        "oidc_callback_path": str,
        # less important ---
        "provider_discovery_uri": str,
        "oidc_logout_redirect_parameter_name": str,
        # Use introspection in API-Bearer mode?
        "use_introspection_on_access_tokens": bool,
        # Rare usages ---
        "client_authn_method": Optional[bool],
        "oidc_logout_query_string_redirect_parameter": Optional[str],
        "oidc_logout_query_string_extra_parameters_dict": Optional[Dict[str, Any]]
        # "client_consumer_config_dict": None,
        # some providers may return even more stuff (...) ---
    },
)


class Provider:
    """
    This is the base `Provider` class that is used to implement common provider configuration patterns. You should not
    use this class directly. Instead, you should but subclass it to implement the configuration logic.
    """

    def __init__(self, *args: Any, op_name: str, **kwargs: Any):
        """
        Parameters:
            op_name (str): the name of the sso provider that you are using
        """

        self.op_name = op_name

        if "provider_discovery_uri" in kwargs:
            self.provider_discovery_uri = kwargs["provider_discovery_uri"]
        else:
            self.provider_discovery_uri = None

        if "oidc_logout_redirect_parameter_name" in kwargs:
            self.oidc_logout_redirect_parameter_name = kwargs[
                "oidc_logout_redirect_parameter_name"
            ]
        else:
            self.oidc_logout_redirect_parameter_name = "post_logout_redirect"

        if "oidc_callback_path" in kwargs:
            self.oidc_callback_path = kwargs["oidc_callback_path"]
        else:
            self.oidc_callback_path = "/oidc-callback/"

    def get_default_config(self) -> ProviderConfig:
        """Get the default configuration settings for this provider.

        This configuration defaults are used to provide default values for OIDCSettings.
        User can override these defaults by playing with OIDCSettings arguments.
        """
        return ProviderConfig(
            # most important ----
            client_id=None,
            client_secret=None,
            oidc_cache_provider_metadata=None,
            oidc_callback_path=self.oidc_callback_path,
            # less important ---
            provider_discovery_uri=self.provider_discovery_uri,
            oidc_logout_redirect_parameter_name=self.oidc_logout_redirect_parameter_name,
            # Use introspection in API-Bearer mode?
            use_introspection_on_access_tokens=self.op_name == "drf",
            # Rare usages ---
            client_authn_method=None,
            oidc_logout_query_string_redirect_parameter=None,
            oidc_logout_query_string_extra_parameters_dict=None,
        )

    @property
    def login_uri_name(self) -> str:
        """
        The login viewname (use in :func:`django:django.urls.reverse` django directive for example) of this configuration
        """
        return f"{self.op_name}-login"

    @property
    def logout_uri_name(self) -> str:
        """
        The logout viewname (use in :func:`django:django.urls.reverse` django directive for example) of this configuration
        """
        return f"{self.op_name}-logout"

    @property
    def callback_uri_name(self) -> str:
        """
        The callback viewname (use in :func:`django:django.urls.reverse` django directive for example) of this configuration
        """
        return f"{self.op_name}-callback"

    def get_urlpatterns(self) -> List[Any]:
        """
        Returns:
            A list of urllpatterns to be included using :func:`django:django.urls.include` in your url configuration
        """
        from django_pyoidc.views import OIDCCallbackView, OIDCLoginView, OIDCLogoutView

        result = [
            path(
                f"{self.oidc_callback_path}",
                OIDCCallbackView.as_view(op_name=self.op_name),
                name=self.callback_uri_name,
            ),
            path(
                f"{self.oidc_callback_path}-login",
                OIDCLoginView.as_view(op_name=self.op_name),
                name=self.login_uri_name,
            ),
            path(
                f"{self.oidc_callback_path}-logout",
                OIDCLogoutView.as_view(op_name=self.op_name),
                name=self.logout_uri_name,
            ),
        ]
        return result



import http.client as http_client
import logging
from urllib.parse import parse_qs, urlparse

import requests
from django.urls import reverse
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.firefox.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions as EC

# from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support.ui import WebDriverWait

from tests.e2e.utils import OIDCE2ELemonLdapNgTestCase

# HTTP debug for requests
http_client.HTTPConnection.debuglevel = 1


class LemonLDAPTestCase(OIDCE2ELemonLdapNgTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        options = Options()
        # options.headless = True
        cls.selenium = WebDriver(options=options)
        # cls.selenium.implicitly_wait(10)

    @classmethod
    def tearDownClass(cls):
        cls.selenium.quit()
        super().tearDownClass()

    def test_00_login_page_redirects_to_lemonldap_sso(self, *args):
        """
        Test that accessing login callback redirects to the SSO server.
        """
        logging.basicConfig()
        logging.getLogger().setLevel(logging.DEBUG)
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True

        client = requests

        print(f"Running Django on {self.live_server_url}")

        login_url = reverse("e2e_test_ll_login_1")
        response = client.get(
            f"{self.live_server_url}{login_url}", allow_redirects=False
        )
        # we should get something like http://localhost:8070/oauth2/authorize
        #   ?client_id=app1
        #   &nonce=YZA...24
        #   &redirect_uri=http%3A%2F%2Ftestserver%2Fcallback
        #   &response_type=code
        #   &scope=openid
        #   &state=fe874.....538e047a0
        self.assertEqual(response.status_code, 302)
        location = response.headers["Location"]
        parsed = urlparse(location)
        qs = parse_qs(parsed.query)
        self.assertEqual(parsed.scheme, "http")
        self.assertEqual(parsed.netloc, "localhost:8070")
        self.assertEqual(parsed.path, "/oauth2/authorize")
        self.assertEqual(qs["client_id"][0], "app1")
        self.assertEqual(qs["redirect_uri"][0], f"{self.live_server_url}/callback-ll-1")
        self.assertEqual(qs["response_type"][0], "code")
        self.assertEqual(qs["scope"][0], "openid")
        self.assertTrue(qs["state"][0])
        self.assertTrue(qs["nonce"][0])

    def _selenium_sso_login(
        self, login_start_url, login_end_url, user, password, active_sso_session=False
    ):
        if not active_sso_session:
            self.selenium.get(login_start_url)
            self.wait.until(EC.url_changes(login_start_url))
            # wait ->
            #   EC.url_changes('a') // anything but a
            #   EC.url_contains('')
            #   EC.url_matches(pattern)
            #   EC.title_contains('')
            #   EC.title_is('')
            #   EC.url_to_be('') // exact

            username_input = self.selenium.find_element(By.NAME, "user")
            username_input.send_keys(user)
            password_input = self.selenium.find_element(By.NAME, "password")
            password_input.send_keys(password)
            self.selenium.find_element(
                By.XPATH, '//form[@id="lform"]//button[contains(@class, "btn-success")]'
            ).click()
        else:
            # current SSO session is still active,
            # so we should redirected directly to success page
            self.selenium.get(login_start_url)
        self.wait.until(EC.url_matches(login_end_url))

    def _selenium_logout(self, end_url):
        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we have the logout link
        self.assertTrue("OIDC-LOGOUT-LINK" in bodyText)
        # click logout
        self.selenium.find_element(By.ID, "oidc-logout-link").click()
        self.wait.until(EC.url_matches(end_url))
        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we are NOT logged in
        self.assertTrue("You are logged out" in bodyText)

    def test_01_selenium_sso_login(self, *args):
        """
        Test a complete working OIDC login.
        """
        timeout = 5
        login_url = reverse("e2e_test_ll_login_1")
        success_url = reverse("e2e_test_ll_success_1")
        start_url = f"{self.live_server_url}{login_url}"
        end_url = f"{self.live_server_url}{success_url}"
        self.wait = WebDriverWait(self.selenium, timeout)
        self._selenium_sso_login(start_url, end_url, "rtyler", "rtyler")

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we are logged in
        self.assertTrue("You are logged in as rtyler@badwolf.org" in bodyText)
        self.assertFalse("You are logged out" in bodyText)
        # Check the session message is shown
        self.assertTrue("message: rtyler@badwolf.org is logged in." in bodyText)

        # check we have the logout link
        self.assertFalse("OIDC-LOGIN-LINK" in bodyText)
        self.assertTrue("OIDC-LOGOUT-LINK" in bodyText)

    def test_02_selenium_sso_login_and_logout(self, *args):
        """
        FIXME : Make this test independant of test #1
        """
        timeout = 5
        login_url = reverse("e2e_test_ll_login_1")
        success_url = reverse("e2e_test_ll_success_1")
        post_logout_url = reverse("e2e_test_ll_logout_done_1")
        start_url = f"{self.live_server_url}{login_url}"
        middle_url = f"{self.live_server_url}{success_url}"
        end_url = f"{self.live_server_url}{post_logout_url}"
        self.wait = WebDriverWait(self.selenium, timeout)

        self._selenium_sso_login(
            start_url, middle_url, "rtyler", "rtyler", active_sso_session=True
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we have the logout link
        self.assertFalse("OIDC-LOGIN-LINK" in bodyText)
        self.assertTrue("OIDC-LOGOUT-LINK" in bodyText)
        # click logout
        self.selenium.find_element(By.ID, "oidc-logout-link").click()

        self.wait.until(EC.url_matches(end_url))

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we are NOT logged in
        print(bodyText)
        self.assertFalse("You are logged in as rtyler@badwolf.org" in bodyText)
        self.assertTrue("You are logged out" in bodyText)

        # Check the new session message is shown
        self.assertTrue("message: post logout view." in bodyText)
        # check we have the login link
        self.assertTrue("OIDC-LOGIN-LINK" in bodyText)
        self.assertFalse("OIDC-LOGOUT-LINK" in bodyText)

    def test_03_selenium_sso_session_with_callbacks(self, *args):
        timeout = 5
        login_url = reverse("e2e_test_ll_login_1")
        success_url = reverse("e2e_test_ll_success_1")
        post_logout_url = reverse("e2e_test_ll_logout_done_1")
        start_url = f"{self.live_server_url}{login_url}"
        middle_url = f"{self.live_server_url}{success_url}"
        end_url = f"{self.live_server_url}{post_logout_url}"
        self.wait = WebDriverWait(self.selenium, timeout)
        # previous test destroyed the SSO session, we need to reconnect
        self._selenium_sso_login(
            start_url, middle_url, "rtyler", "rtyler", active_sso_session=False
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # print(bodyText)

        # Check the session message is shown
        self.assertTrue("message: rtyler@badwolf.org is logged in." in bodyText)
        # Check the CALLBACK session message is also shown
        self.assertTrue("login_callback for : rtyler" in bodyText)

        # check we have the logout link
        self.assertTrue("OIDC-LOGOUT-LINK" in bodyText)
        # click logout
        self.selenium.find_element(By.ID, "oidc-logout-link").click()

        self.wait.until(EC.url_matches(end_url))

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we are NOT logged in
        self.assertTrue("You are logged out" in bodyText)

        # Check the new session message is shown
        self.assertTrue("message: post logout view." in bodyText)
        # Check the logout callback message is shown
        self.assertTrue("Logout Callback for rtyler@badwolf.org." in bodyText)

    # def test_04_selenium_sso_failed_login(self, *args):
    #    """
    #    Test a failed SSO login (bad client)
    #    """
    #    timeout = 30
    #    login_url = reverse("e2e_test_ll_login_2")
    #    start_url = f"{self.live_server_url}{login_url}"
    #    wait = WebDriverWait(self.selenium, timeout)
    #    # wait ->
    #    #   EC.url_changes('a') // anything but a
    #    #   EC.url_contains('')
    #    #   EC.url_matches(pattern)
    #    #   EC.title_contains('')
    #    #   EC.title_is('')
    #    #   EC.url_to_be('') // exact
    # (...)



from typing import Any, Dict

from django.contrib.auth import get_user_model
from django.core.exceptions import SuspiciousOperation

from django_pyoidc.exceptions import ClaimNotFoundError
from django_pyoidc.utils import extract_claim_from_tokens


def get_user_by_email(tokens: Dict[str, Any]) -> Any:
    User = get_user_model()

    username = None
    preferred_username = None
    email = None
    client_host = None
    client_address = None
    client_id = None
    django_username = None

    try:
        email = extract_claim_from_tokens("email", tokens)
    except ClaimNotFoundError:
        pass

    try:
        preferred_username = extract_claim_from_tokens("preferred_username", tokens)
    except ClaimNotFoundError:
        pass

    try:
        username = extract_claim_from_tokens("username", tokens)
    except ClaimNotFoundError:
        pass

    if preferred_username is None:
        if username is None:
            if email:
                django_username = email
            else:
                raise SuspiciousOperation(
                    "Cannot extract username or email from available OIDC tokens."
                )
        else:
            django_username = username
    else:
        django_username = preferred_username

    # Currently client Credential logins in Keycloak adds these mappers in access tokens.
    # we can use that to detect M2M accounts.
    # TODO: check if any other mapper can/should be used for other providers
    try:
        client_host = extract_claim_from_tokens("clientHost", tokens)
    except ClaimNotFoundError:
        pass
    try:
        client_address = extract_claim_from_tokens("clientAddress", tokens)
    except ClaimNotFoundError:
        pass

    if email is None and client_host is not None and client_address is not None:
        # that's a M2M connexion in grant=client_credential mode
        try:
            client_id = extract_claim_from_tokens("client_id", tokens)
        except ClaimNotFoundError:
            client_id = django_username
        # Build a fake email for the service accounts
        email = f"{client_id}@localhost.lan"

    user, created = User.objects.get_or_create(
        email=email,
        username=django_username,
    )
    if hasattr(user, "backend"):
        user.backend = "django.contrib.auth.backends.ModelBackend"
    return user



import logging
from functools import lru_cache
from importlib import import_module
from typing import Any, Dict, List, Optional, TypedDict, TypeVar, Union

from django.conf import settings as django_settings
from django.urls import reverse_lazy

from django_pyoidc.exceptions import InvalidOIDCConfigurationException

logger = logging.getLogger(__name__)

T = TypeVar("T")

TypedOidcSettings = TypedDict(
    "TypedOidcSettings",
    {
        "cache_django_backend": str,
        "oidc_cache_provider_metadata": bool,
        "oidc_cache_provider_metadata_ttl": int,
        "use_introspection_on_access_tokens": bool,
    },
)

OidcSettingValue = Union[bool, int, str, List[str]]


class OIDCSettings:

    GLOBAL_SETTINGS: TypedOidcSettings = {
        "cache_django_backend": "default",
        "oidc_cache_provider_metadata": False,
        "oidc_cache_provider_metadata_ttl": 120,
        "use_introspection_on_access_tokens": True,
    }

    def __repr__(self) -> str:
        repr_str = f"Oidc Settings for {self.op_name}"
        for key, val in self.OP_SETTINGS.items():
            if val is None:
                repr_str += f"\n + {key}: None"
            else:
                repr_str += f"\n + {key}: {val}"
        return repr_str

    def __init__(self, op_name: str):
        """
           Parameters:
               op_name (str): the name of the sso provider that you are using
        Other settings are loaded from Django setting DJANGO_PYOIDC, prefixed by the op_name
        It may for example contain:
               client_secret (str): the OIDC client secret
               client_id (str): the OIDC client ID
               provider_discovery_uri (str): URL of the SSO server (the .well-known/openid-configuration part is added to this path).
                Some providers like the keycloak provider can instead generate this settings by combining 'keycloak_base_uri' (str) and
                'keycloak_realm' (str) settings.
               oidc_callback_path (str): the path used to call this library during the login round-trips, the default is "/oidc-callback/".
               callback_uri_name (str): the route giving the path for oidc_callback_path that you can use instead of oidc_callback_path
               post_logout_redirect_uri (str): the URI where a user should be redirected to on logout success
               post_login_uri_failure (str): the URI where a user should be redirected to on login failure
               post_login_uri_success (str): the URI a user should be redirected to on login success if no redirection url where provided
               login_redirection_requires_https (bool): set to True to disallow redirecting user to non-https uri on login success
               login_uris_redirect_allowed_hosts(:obj:`list`) : A list of allowed domains that can be redirected to.
                 A good idea is to this to use :setting:`ALLOWED_HOSTS <django:ALLOWED_HOSTS>`.
                 See :ref:`Redirect the user after login` for more details.
               oidc_cache_provider_metadata (bool): default to False; if True calls to the provider_discovery_uri will be cached,
                removing a lot of HTTP traffic. The provider metadata is the same for all your users, so when you havce a lot of
                concurrent OIDC related operations this cache can be useful even with a short duration.
               oidc_cache_provider_metadata_ttl (int): validity of the metadata cache in seconds, default is 120 (2 minutes).
                you can use a long TTL (you know the SSO metadata does not move a lot) or a shorter one (microcache).
               cache_django_backend(:obj:`str`, optional): Defaults to 'default'. The cache backend that should be used to store
                 this provider sessions. Take a look at :ref:`Cache Management`
               hook_user_login (str): path to a function hook to be run after successful login.
               hook_user_logout (str):  path to a function hook to be run during logout(before local session removal and redirection to SSO
                remote logout).
               hook_validate_access_token (str):  path to a function hook to extract access tokens claims from the raw jwt.
                this is not used if 'use_introspection_on_access_tokens' is True
               use_introspection_on_access_tokens (bool): extract access tokens claims by sending the access token to the sso server on
                the introspection route. This delegates validation of the token to the SSO server. If you do not use hook_validate_access_token
                or use_introspection_on_access_tokens you will just have the raw jwt for the access token, that you can use to send HTTP queries
                on behalf of the user.
        """

        self.op_name = op_name
        if self.op_name not in django_settings.DJANGO_PYOIDC:
            raise InvalidOIDCConfigurationException(
                f"{self.op_name} provider name must be configured in DJANGO_PYOIDC settings."
            )

        op_definition = {
            k.lower(): v for k, v in django_settings.DJANGO_PYOIDC[self.op_name].items()
        }

        # fix potential bad settings declaration (or aliases)
        op_definition = self._fix_settings(op_definition)
        if "provider_class" in op_definition:
            provider_class = op_definition["provider_class"]
            # allow usage of simple names like "keycloak" instead of "django_pyoidc.providers.keycloak"
            if "." not in provider_class:
                provider_class = f"django_pyoidc.providers.{provider_class}"
        else:
            provider_class = "django_pyoidc.providers.Provider"

        provider_module_path, provider_class = provider_class.rsplit(".", 1)
        provider_real_class = getattr(
            import_module(provider_module_path), provider_class
        )

        # This call can fail if required attributes are not set
        provider = provider_real_class(op_name=self.op_name, **op_definition)

        # Init a local final operator settings with user given values
        self.OP_SETTINGS = op_definition
        # get some defaults and variation set by the provider
        # For example it could be a newly computed value (provider_discovery_uri from uri and realm for keycloak)
        # or some defaults altered
        provider_default_settings = provider.get_default_config()
        # Then merge the two, so for all settings we have, with priority
        #  * user defined specific values (if not empty)
        #  * provider computed or default value (if not empty)
        #  * then later if a get() is made on this value we'll have the globals applied
        for key, val in provider_default_settings.items():
            key = key.lower()
            if key not in self.OP_SETTINGS or self.OP_SETTINGS[key] is None:
                self.OP_SETTINGS[key] = val

        self.OP_SETTINGS["op_name"] = self.op_name
        self._validate_settings()

    def _fix_settings(self, op_definition: Dict[str, Any]) -> Dict[str, Any]:
        """Workarounds over specific settings and aliases."""

        # pyoidc wants the discovery uri WITHOUT the well-known part '.well-known/openid-configuration'
        if (
            "provider_discovery_uri" in op_definition
            and op_definition["provider_discovery_uri"]
        ):
            discovery = op_definition["provider_discovery_uri"]
            extra_string = ".well-known/openid-configuration"
            if discovery.endswith(extra_string):
                discovery = discovery[: -len(extra_string)]
                op_definition["provider_discovery_uri"] = discovery
            extra_string = ".well-known/openid-configuration/"
            if discovery.endswith(extra_string):
                discovery = discovery[: -len(extra_string)]
                op_definition["provider_discovery_uri"] = discovery
            extra_string = "/"
            if discovery.endswith(extra_string):
                discovery = discovery[: -len(extra_string)]
                op_definition["provider_discovery_uri"] = discovery

        # Special path manipulations
        if "oidc_callback_path" in op_definition:
            op_definition["oidc_callback_path"] = op_definition["oidc_callback_path"]
        if "callback_uri_name" in op_definition:
            op_definition["oidc_callback_path"] = reverse_lazy(
                op_definition["callback_uri_name"]
            )
            del op_definition["callback_uri_name"]
        # else: do not set defaults.
        # The Provider objet will define a defaut callback path if not set.

        # allow simpler names
        # * "logout_redirect" for "post_logout_redirect_uri"
        # * "failure_redirect" for "post_login_uri_failure"
        # * "success_redirect" for "post_login_uri_success"
        # * "redirect_requires_https" for "login_redirection_requires_https"
        if "post_logout_redirect_uri" not in op_definition:
            if "logout_redirect" in op_definition:
                op_definition["post_logout_redirect_uri"] = op_definition[
                    "logout_redirect"
                ]
                del op_definition["logout_redirect"]
            else:
                op_definition["post_logout_redirect_uri"] = "/"

        if "post_login_uri_failure" not in op_definition:
            if "failure_redirect" in op_definition:
                op_definition["post_login_uri_failure"] = op_definition[
                    "failure_redirect"
                ]
                del op_definition["failure_redirect"]
            else:
                op_definition["post_login_uri_failure"] = "/"

        if "post_login_uri_success" not in op_definition:
            if "success_redirect" in op_definition:
                op_definition["post_login_uri_success"] = op_definition[
                    "success_redirect"
                ]
                del op_definition["success_redirect"]
            else:
                op_definition["post_login_uri_success"] = "/"

        if "login_redirection_requires_https" not in op_definition:
            if "redirect_requires_https" in op_definition:
                op_definition["login_redirection_requires_https"] = op_definition[
                    "redirect_requires_https"
                ]
                del op_definition["redirect_requires_https"]
            else:
                op_definition["login_redirection_requires_https"] = True

        return op_definition

    def _validate_settings(self) -> None:
        if (
            "hook_validate_access_token" in self.OP_SETTINGS
            and "use_introspection_on_access_tokens" in self.OP_SETTINGS
            and self.OP_SETTINGS["use_introspection_on_access_tokens"]
        ):
            raise InvalidOIDCConfigurationException(
                "You cannot define hook_validate_access_token if you use use_introspection_on_access_tokens."
            )

        # client_id is required
        if "client_id" not in self.OP_SETTINGS or self.OP_SETTINGS["client_id"] is None:
            raise InvalidOIDCConfigurationException(
                f"Provider definition does not contain any 'client_id' entry. Check your DJANGO_PYOIDC['{self.op_name}'] settings."
            )
        # we do not enforce client_secret (in case someone wrongly use a public client)
        if (
            "client_secret" not in self.OP_SETTINGS
            or self.OP_SETTINGS["client_secret"] is None
        ):
            logger.warning(
                f"OIDC settings for {self.op_name} has no client_secret. You are maybe using a public OIDC client, you should not."
            )

    def set(self, key: str, value: Optional[OidcSettingValue] = None) -> None:
        self.OP_SETTINGS[key] = value

    def get(
        self, name: str, default: Optional[Union[OidcSettingValue, T]] = None
    ) -> Optional[Union[OidcSettingValue, T]]:
        "Get attr value for op or global, given last arg is the default value if None."
        res = self._get_attr(name)
        if res is None:
            return default
        return res

    def _get_attr(self, key: str) -> Optional[OidcSettingValue]:
        """Retrieve attr, if op value is None a check on globals is made.

        Note that op value is already a computation of provider defaults and user defined settings.
        """
        key = key.lower()
        if key in self.OP_SETTINGS and self.OP_SETTINGS[key] is not None:
            return self.OP_SETTINGS[key]
        else:
            if key in self.GLOBAL_SETTINGS.keys():
                return self.GLOBAL_SETTINGS[key]  # type: ignore[literal-required] #  we check that the key is available the line before
        return None


class OIDCSettingsFactory:
    @classmethod
    @lru_cache(maxsize=10)
    def get(cls, op_name: str) -> OIDCSettings:
        """
        lru_cache will return a singleton for each argument value.
        So this is a memoized function.
        """
        settings = OIDCSettings(op_name=op_name)
        return settings



#!/usr/bin/env python
import os
import sys

import django
from django.conf import settings
from django.test.utils import get_runner

if __name__ == "__main__":
    os.environ["DJANGO_SETTINGS_MODULE"] = "tests.e2e.settings"
    django.setup()
    E2eTestRunner = get_runner(settings)
    test_runner = E2eTestRunner()
    failures = test_runner.run_tests(["tests/e2e"])
    sys.exit(bool(failures))



from .keycloak import KeycloakProvider  # noqa
from .keycloak_10 import Keycloak10Provider  # noqa
from .keycloak_17 import Keycloak17Provider  # noqa
from .keycloak_18 import Keycloak18Provider  # noqa
from .lemonldapng import LemonLDAPngProvider  # noqa
from .lemonldapng2 import LemonLDAPng2Provider  # noqa
from .provider import Provider  # noqa



from importlib import import_module
from unittest import mock

from django.conf import settings
from django.test import override_settings

from django_pyoidc.client import OIDCClient
from django_pyoidc.exceptions import InvalidOIDCConfigurationException
from tests.utils import OIDCTestCase

SessionStore = import_module(settings.SESSION_ENGINE).SessionStore


class SettingsTestCase(OIDCTestCase):
    @override_settings(
        DJANGO_PYOIDC={
            "lib_123": {
                "oidc_cache_provider_metadata": False,
                "oidc_cache_provider_metadata_ttl": 75,
                "client_id": "foo",
                "client_secret": "secret_app_foo",
                "cache_django_backend": "default",
                "provider_class": "django_pyoidc.providers.KeycloakProvider",
                "keycloak_base_uri": "http://sso_tutut",
                "keycloak_realm": "realm_foo",
                "oidc_callback_path": "/callback-foo-abc",
                "login_uris_redirect_allowed_hosts": ["foo", "bar"],
                "login_redirection_requires_https": True,
                "post_login_uri_success": "/abc-123",
                "post_login_uri_failure": "/def-456",
                "post_logout_redirect_uri": "/ghj-789",
                "hook_user_login": "tests.e2e.test_app.callback:login_callback",
                "hook_user_logout": "tests.e2e.test_app.callback:logout_callback",
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_extracted_from_django_pyoidc_settings(
        self, mocked_provider_config, *args
    ):
        """
        Test that definitions of DJANGO_PYOIDC in settings can be retrieved in client settings.
        """
        sso_client = OIDCClient(op_name="lib_123")
        settings = sso_client.get_settings()
        self.assertEqual(settings.get("client_id"), "foo")
        self.assertEqual(settings.get("client_secret"), "secret_app_foo")
        self.assertEqual(
            settings.get("provider_discovery_uri"), "http://sso_tutut/realms/realm_foo"
        )
        self.assertEqual(settings.get("cache_django_backend"), "default")
        self.assertEqual(settings.get("oidc_cache_provider_metadata"), False)
        self.assertEqual(settings.get("oidc_cache_provider_metadata_ttl"), 75)
        self.assertEqual(settings.get("oidc_callback_path"), "/callback-foo-abc")
        self.assertEqual(
            settings.get("login_uris_redirect_allowed_hosts"), ["foo", "bar"]
        )
        self.assertEqual(settings.get("login_redirection_requires_https"), True)
        self.assertEqual(settings.get("post_login_uri_success"), "/abc-123")
        self.assertEqual(settings.get("post_login_uri_failure"), "/def-456")
        self.assertEqual(settings.get("post_logout_redirect_uri"), "/ghj-789")
        self.assertEqual(
            settings.get("hook_user_login"),
            "tests.e2e.test_app.callback:login_callback",
        )
        self.assertEqual(
            settings.get("hook_user_logout"),
            "tests.e2e.test_app.callback:logout_callback",
        )

    @override_settings(
        DJANGO_PYOIDC={
            "lib_238": {
                "oidc_cache_provider_metadata": False,
                "client_id": "foo2",
                "provider_discovery_uri": "http://foo",
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_defaults(self, mocked_provider_config, *args):
        """
        Test that minimal definitions of DJANGO_PYOIDC still get some default settings.
        """
        sso_client = OIDCClient(op_name="lib_238")
        settings = sso_client.get_settings()
        self.assertEqual(settings.get("client_id"), "foo2")
        # useful if bad public client is used
        self.assertEqual(settings.get("client_secret"), None)
        self.assertEqual(settings.get("provider_discovery_uri"), "http://foo")
        self.assertEqual(settings.get("cache_django_backend"), "default")
        self.assertEqual(settings.get("oidc_cache_provider_metadata"), False)
        self.assertEqual(settings.get("oidc_cache_provider_metadata_ttl"), 120)
        self.assertEqual(settings.get("login_redirection_requires_https"), True)
        self.assertEqual(settings.get("post_login_uri_success"), "/")
        self.assertEqual(settings.get("post_login_uri_failure"), "/")
        self.assertEqual(settings.get("post_logout_redirect_uri"), "/")
        self.assertEqual(settings.get("oidc_callback_path"), "/oidc-callback/")
        self.assertEqual(settings.get("login_uris_redirect_allowed_hosts"), None)
        self.assertEqual(settings.get("hook_user_login"), None)
        self.assertEqual(settings.get("hook_user_logout"), None)
        self.assertEqual(settings.get("hook_validate_access_token"), None)
        self.assertEqual(settings.get("use_introspection_on_access_tokens"), False)

    @override_settings(
        DJANGO_PYOIDC={
            "lib_314": {
                "oidc_cache_provider_metadata": False,
                "client_id": "foo3",
                "client_secret": "secret",
                "provider_discovery_uri": "http://foo",
                "callback_uri_name": "my_test_callback",
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_callback_uri_name_to_oidc_callback_path(
        self, mocked_provider_config, *args
    ):
        """
        Test that we can give a named route in callback_uri_name instead of giving oidc_callback_path.
        """
        sso_client = OIDCClient(op_name="lib_314")
        settings = sso_client.get_settings()
        self.assertEqual(settings.get("callback_uri_name"), None)
        self.assertEqual(settings.get("oidc_callback_path"), "/callback-xyz/")

    @override_settings(
        DJANGO_PYOIDC={
            "lib_315": {
                "oidc_cache_provider_metadata": False,
                "client_id": "foo3",
                "client_secret": "secret",
                "provider_discovery_uri": "http://foo",
                "CALLBACK_URI_NAME": "my_test_callback",
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_callback_uri_name_to_oidc_callback_path_upper(
        self, mocked_provider_config, *args
    ):
        """
        Test that we can give a named route in callback_uri_name instead of giving oidc_callback_path.
        """
        sso_client = OIDCClient(op_name="lib_315")
        settings = sso_client.get_settings()
        self.assertEqual(settings.get("callback_uri_name"), None)
        self.assertEqual(settings.get("CALLBACK_URI_NAME"), None)
        self.assertEqual(settings.get("oidc_callback_path"), "/callback-xyz/")

    @override_settings(
        DJANGO_PYOIDC={
            "lib_318": {
                "oidc_cache_provider_metadata": False,
                "client_id": "foo4",
                "client_secret": "secret",
                "provider_discovery_uri": "http://foo",
                "logout_redirect": "/zorg-1",
                "failure_redirect": "/zorg-2",
                "success_redirect": "/zorg-3",
                "redirect_requires_https": False,
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_aliases(self, mocked_provider_config, *args):
        """
        Test that minimal definitions of DJANGO_PYOIDC still get some default settings.
        """
        sso_client = OIDCClient(op_name="lib_318")
        settings = sso_client.get_settings()
        # Aliases
        # logout_redirect -> post_logout_redirect_uri
        self.assertEqual(settings.get("logout_redirect"), None)
        self.assertEqual(settings.get("post_logout_redirect_uri"), "/zorg-1")
        # failure_redirect -> post_login_uri_failure
        self.assertEqual(settings.get("failure_redirect"), None)
        self.assertEqual(settings.get("post_login_uri_failure"), "/zorg-2")
        # success_redirect -> post_login_uri_success
        self.assertEqual(settings.get("success_redirect"), None)
        self.assertEqual(settings.get("post_login_uri_success"), "/zorg-3")
        # redirect_requires_https -> login_redirection_requires_https
        self.assertEqual(settings.get("redirect_requires_https"), None)
        self.assertEqual(settings.get("login_redirection_requires_https"), False)

    @override_settings(
        DJANGO_PYOIDC={
            "lib_319": {
                "oidc_cache_provider_metadata": False,
                "client_id": "foo4",
                "client_secret": "secret",
                "provider_discovery_uri": "http://foo",
                "LOGOUT_REDIRECT": "/zorg-1",
                "FAILURE_REDIRECT": "/zorg-2",
                "SUCCESS_REDIRECT": "/zorg-3",
                "REDIRECT_REQUIRES_HTTPS": False,
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_aliases_upper(self, mocked_provider_config, *args):
        """
        Test that minimal definitions of DJANGO_PYOIDC still get some default settings.
        """
        sso_client = OIDCClient(op_name="lib_319")
        settings = sso_client.get_settings()
        # Aliases
        # logout_redirect -> post_logout_redirect_uri
        self.assertEqual(settings.get("LOGOUT_REDIRECT"), None)
        self.assertEqual(settings.get("logout_redirect"), None)
        self.assertEqual(settings.get("post_logout_redirect_uri"), "/zorg-1")
        # failure_redirect -> post_login_uri_failure
        self.assertEqual(settings.get("failure_redirect"), None)
        self.assertEqual(settings.get("FAILURE_REDIRECT"), None)
        self.assertEqual(settings.get("post_login_uri_failure"), "/zorg-2")
        # success_redirect -> post_login_uri_success
        self.assertEqual(settings.get("success_redirect"), None)
        self.assertEqual(settings.get("SUCCESS_REDIRECT"), None)
        self.assertEqual(settings.get("post_login_uri_success"), "/zorg-3")
        # redirect_requires_https -> login_redirection_requires_https
        self.assertEqual(settings.get("redirect_requires_https"), None)
        self.assertEqual(settings.get("REDIRECT_REQUIRES_HTTPS"), None)
        self.assertEqual(settings.get("login_redirection_requires_https"), False)

    @override_settings(
        DJANGO_PYOIDC={
            "lib_371": {
                "oidc_cache_provider_metadata": False,
                "client_id": "foo3",
                "client_secret": "secret",
                "provider_discovery_uri": "http://foo",
                "callback_uri_name": "my_test_callback",
                "use_introspection_on_access_tokens": True,
                "hook_validate_access_token": "tests.e2e.test_app.callback:hook_validate_access_token",
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_no_hook_validate_access_token_if_use_introspection_on_access_tokens(
        self, mocked_provider_config, *args
    ):
        """
        Test that we prevent setting using both use_introspection_on_access_tokens and hook_validate_access_token.
        """
        with self.assertRaises(InvalidOIDCConfigurationException) as context:
            OIDCClient(op_name="lib_371")
        self.assertTrue(
            "You cannot define hook_validate_access_token if you use use_introspection_on_access_tokens."
            in context.exception.__repr__()
        )

    @override_settings(
        DJANGO_PYOIDC={
            "lib_547": {
                "oidc_cache_provider_metadata": False,
                "client_secret": "secret_app_foo2",
                "provider_discovery_uri": "http://foo",
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_missing_client_id(self, mocked_provider_config, *args):
        """
        Test that missing client_id will fail.
        """
        with self.assertRaises(InvalidOIDCConfigurationException) as context:
            OIDCClient(op_name="lib_547")
        self.assertTrue(
            "Provider definition does not contain any 'client_id' entry."
            in context.exception.__repr__()
        )

    @override_settings(
        DJANGO_PYOIDC={
            "lib_548": {
                "oidc_cache_provider_metadata": False,
                "client_id": "lib_548",
                "client_secret": "secret",
            },
            "lib_549": {
                "oidc_cache_provider_metadata": False,
                "client_id": "lib_549",
                "client_secret": "secret",
                "provider_class": "django_pyoidc.providers.KeycloakProvider",
                "keycloak_base_uri": "http://sso_tutut",
            },
            "lib_550": {
                "oidc_cache_provider_metadata": False,
                "client_id": "lib_550",
                "client_secret": "secret",
                "provider_class": "django_pyoidc.providers.KeycloakProvider",
                "keycloak_realm": "toto",
            },
            "lib_551": {
                "oidc_cache_provider_metadata": False,
                "client_id": "lib_551",
                "client_secret": "secret",
                "provider_class": "django_pyoidc.providers.KeycloakProvider",
                "provider_discovery_uri": "http://uvw/xyz/abc/",
            },
            "lib_552": {
                "oidc_cache_provider_metadata": False,
                "client_id": "lib_552",
                "client_secret": "secret",
                "provider_class": "django_pyoidc.providers.KeycloakProvider",
                "provider_discovery_uri": "http://uvw/xyz/realms/",
            },
            "lib_553": {
                "oidc_cache_provider_metadata": False,
                "client_id": "lib_552",
                "client_secret": "secret",
                "provider_class": "django_pyoidc.providers.KeycloakProvider",
                "provider_discovery_uri": "http://uvw/xyz/realms/foo/bar",
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_missing_provider_discovery_uri(
        self, mocked_provider_config, *args
    ):
        """
        Test that missing provider_discovery_uri (or alternatives) will fail.
        """
        with self.assertRaises(InvalidOIDCConfigurationException) as context:
            OIDCClient(op_name="lib_548")
        self.assertTrue(
            "No provider discovery uri provided." in context.exception.__repr__()
        )
        with self.assertRaises(TypeError) as context:
            OIDCClient(op_name="lib_549")
        self.assertTrue(
            "Keycloak10Provider requires keycloak_base_uri and keycloak_realm or provider_discovery_uri."
            in context.exception.__repr__()
        )
        with self.assertRaises(TypeError) as context:
            OIDCClient(op_name="lib_550")
        self.assertTrue(
            "Keycloak10Provider requires keycloak_base_uri and keycloak_realm or provider_discovery_uri."
            in context.exception.__repr__()
        )
        with self.assertRaises(RuntimeError) as context:
            OIDCClient(op_name="lib_551")
        self.assertTrue(
            "Provided 'provider_discovery_uri' url is not a valid Keycloak metadata url, it does not contains /realms/."
            in context.exception.__repr__()
        )
        with self.assertRaises(RuntimeError) as context:
            OIDCClient(op_name="lib_552")
        self.assertTrue(
            "Provided 'provider_discovery_uri' url is not a valid Keycloak metadata url, it does not contains /realms/."
            in context.exception.__repr__()
        )
        with self.assertRaises(RuntimeError) as context:
            OIDCClient(op_name="lib_553")
        self.assertTrue(
            "Cannot extract the keycloak realm from the provided url."
            in context.exception.__repr__()
        )

    @override_settings(
        DJANGO_PYOIDC={
            "lib_612": {
                "oidc_cache_provider_metadata": False,
                "client_id": "lib_612",
                "client_secret": "secret",
                "provider_class": "django_pyoidc.providers.KeycloakProvider",
                "keycloak_base_uri": "http://abc/def",
                "keycloak_realm": "ghj",
            },
            "lib_613": {
                "oidc_cache_provider_metadata": False,
                "client_id": "lib_613",
                "client_secret": "secret",
                "provider_class": "django_pyoidc.providers.KeycloakProvider",
                "provider_discovery_uri": "http://lmn/opq/realms/rst",
            },
            "lib_614": {
                "oidc_cache_provider_metadata": False,
                "client_id": "lib_613",
                "client_secret": "secret",
                "provider_class": "django_pyoidc.providers.KeycloakProvider",
                "provider_discovery_uri": "http://uvw/xyz/realms/abc/.well-known/openid-configuration",
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_keycloak_provider_can_generate_provider_discovery_uri_or_not(
        self, mocked_provider_config, *args
    ):
        """
        Test that missing provider_discovery_uri (or alternatives) will fail.
        """
        sso_client = OIDCClient(op_name="lib_612")
        settings = sso_client.get_settings()
        self.assertEqual(
            settings.get("provider_discovery_uri"), "http://abc/def/realms/ghj"
        )
        sso_client = OIDCClient(op_name="lib_613")
        settings = sso_client.get_settings()
        self.assertEqual(
            settings.get("provider_discovery_uri"), "http://lmn/opq/realms/rst"
        )
        sso_client = OIDCClient(op_name="lib_614")
        settings = sso_client.get_settings()
        self.assertEqual(
            settings.get("provider_discovery_uri"), "http://uvw/xyz/realms/abc"
        )

    @override_settings(
        DJANGO_PYOIDC={
            "lib_885": {
                "OIDC_CACHE_PROVIDER_METADATA": False,
                "CLIENT_ID": "foo2",
                "CLIENT_SECRET": "secret_app_foo2",
                "CACHE_DJANGO_BACKEND": "default",
                "PROVIDER_DISCOVERY_URI": "http://localhost:8080/auth/realms/stuff",
                "OIDC_CALLBACK_PATH": "/callback-foo-def",
                "LOGIN_URIS_REDIRECT_ALLOWED_HOSTS": ["foo2", "bar2"],
                "LOGIN_REDIRECTION_REQUIRES_HTTPS": True,
                "POST_LOGIN_URI_SUCCESS": "/abc-123-2",
                "POST_LOGIN_URI_FAILURE": "/def-456-2",
                "POST_LOGOUT_REDIRECT_URI": "/ghj-789-2",
                "HOOK_USER_LOGIN": "tests.e2e.test_app.callback:login_callback",
                "HOOK_USER_LOGOUT": "tests.e2e.test_app.callback:logout_callback",
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_extracted_from_django_pyoidc_settings_upper(
        self, mocked_provider_config, *args
    ):
        """
        Test that definitions of DJANGO_PYOIDC in settings can be retrieved in client settings.

        Here testing that cas is not taken into account.
        """
        sso_client = OIDCClient(op_name="lib_885")
        settings = sso_client.get_settings()
        self.assertEqual(settings.get("client_id"), "foo2")
        self.assertEqual(settings.get("client_secret"), "secret_app_foo2")
        self.assertEqual(
            settings.get("provider_discovery_uri"),
            "http://localhost:8080/auth/realms/stuff",
        )
        self.assertEqual(settings.get("cache_django_backend"), "default")
        self.assertEqual(settings.get("oidc_cache_provider_metadata"), False)
        self.assertEqual(settings.get("oidc_callback_path"), "/callback-foo-def")
        self.assertEqual(
            settings.get("login_uris_redirect_allowed_hosts"), ["foo2", "bar2"]
        )
        self.assertEqual(settings.get("login_redirection_requires_https"), True)
        self.assertEqual(settings.get("post_login_uri_success"), "/abc-123-2")
        self.assertEqual(settings.get("post_login_uri_failure"), "/def-456-2")
        self.assertEqual(settings.get("post_logout_redirect_uri"), "/ghj-789-2")
        self.assertEqual(
            settings.get("hook_user_login"),
            "tests.e2e.test_app.callback:login_callback",
        )
        self.assertEqual(
            settings.get("hook_user_logout"),
            "tests.e2e.test_app.callback:logout_callback",
        )

    @override_settings(
        DJANGO_PYOIDC={
            "lib_901": {
                "oidc_cache_provider_metadata": False,
                "client_id": "zorg",
                "client_secret": "--",
                "provider_discovery_uri": "http://foobar/zorg/.well-known/openid-configuration/",
            },
            "lib_902": {
                "oidc_cache_provider_metadata": False,
                "client_id": "zorg",
                "client_secret": "--",
                "provider_discovery_uri": "http://foobar/zorg2/.well-known/openid-configuration",
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_well_known_part_removed_from_provider_path(
        self, mocked_provider_config, *args
    ):
        """
        Test giving .well-known/openid-configuration/ paths for provider config paths is supported.
        """
        sso_client = OIDCClient(op_name="lib_901")
        settings = sso_client.get_settings()
        self.assertEqual(settings.get("provider_discovery_uri"), "http://foobar/zorg")
        sso_client = OIDCClient(op_name="lib_902")
        settings = sso_client.get_settings()
        self.assertEqual(settings.get("provider_discovery_uri"), "http://foobar/zorg2")

    @override_settings(
        DJANGO_PYOIDC={
            "lib_865": {
                "oidc_cache_provider_metadata": False,
                "client_id": "eorg",
                "client_secret": "--",
                "provider_discovery_uri": "http://foo",
            },
        },
    )
    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_settings_globals(self, mocked_provider_config, *args):
        """
        Test that some globale settings are defined.
        """
        sso_client = OIDCClient(op_name="lib_865")
        settings = sso_client.get_settings()
        self.assertEqual(settings.get("CACHE_DJANGO_BACKEND"), "default")
        self.assertEqual(settings.get("cache_django_backend"), "default")
        self.assertEqual(settings.get("OIDC_CACHE_PROVIDER_METADATA"), False)
        self.assertEqual(settings.get("oidc_cache_provider_metadata"), False)
        self.assertEqual(settings.get("OIDC_CACHE_PROVIDER_METADATA_TTL"), 120)
        self.assertEqual(settings.get("oidc_cache_provider_metadata_ttl"), 120)
        self.assertEqual(settings.get("USE_INTROSPECTION_ON_ACCESS_TOKENS"), False)
        self.assertEqual(settings.get("use_introspection_on_access_tokens"), False)

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_client_use_introspection_on_access_tokens_defaults_to_true_if_name_drf(
        self, *args
    ):
        """
        Test that some globale settings are defined.
        """
        sso_client = OIDCClient(op_name="drf")
        settings = sso_client.get_settings()
        self.assertEqual(settings.get("USE_INTROSPECTION_ON_ACCESS_TOKENS"), True)
        self.assertEqual(settings.get("use_introspection_on_access_tokens"), True)



from typing_extensions import override

from django_pyoidc.providers.keycloak_17 import Keycloak17Provider
from django_pyoidc.providers.provider import ProviderConfig


class Keycloak18Provider(Keycloak17Provider):
    """
    Provide Django settings/urlconf based on keycloak behaviour (v18)
    """

    @override
    def get_default_config(self) -> ProviderConfig:
        result = super().get_default_config()
        # logout redirection query string parameter name altered, from redirect_uri to post_logout_redirect_uri
        result[
            "oidc_logout_query_string_redirect_parameter"
        ] = "post_logout_redirect_uri"
        return result



from unittest import mock

from django.conf import settings
from django.urls import reverse

from tests.utils import OIDCTestCase


class UrlsTestCase(OIDCTestCase):
    def test_reverse_urls_are_defined(self, *args):
        """
        Test that urls defined in our test settings are working.
        """
        self.assertEqual(
            reverse("test_blogout"),
            "/back_channel_logout-xyz/",
        )
        self.assertEqual(
            reverse("test_login"),
            "/login-xyz/",
        )
        self.assertEqual(
            reverse("test_logout"),
            "/logout-xyz/",
        )
        self.assertEqual(
            reverse("test_logout"),
            "/logout-xyz/",
        )

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_callback_path_with_uri_name(self, *args):
        """
        Test that using callback_uri_name we have the right callback path.

        "sso1" definition is not using oidc_callback_path but a named route
        with callback_uri_name instead.
        """
        host = settings.DJANGO_PYOIDC["sso1"]["login_uris_redirect_allowed_hosts"][0]
        response = self.client.get(
            reverse("test_login"),
            data={"next": f"https://{host}/foo/bar"},
            SERVER_NAME=host,
        )
        self.assertEqual(response.status_code, 302)
        location = response.headers["Location"]
        elements = location.split("?")
        query_string = elements[1]
        arguments = query_string.split("&")
        self.assertIn("client_id=1", arguments)
        self.assertIn(f"redirect_uri=http%3A%2F%2F{host}%2Fcallback-xyz%2F", arguments)

    @mock.patch("django_pyoidc.client.Consumer.provider_config")
    def test_callback_path_with_oidc_callback_path(self, *args):
        """
        Test that using only oidc_callback_path we have the right callback path.

        "sso2" definition does not contain callback_uri_name but oidc_callback_path
        instead.
        """
        host = settings.DJANGO_PYOIDC["sso2"]["login_uris_redirect_allowed_hosts"][0]
        response = self.client.get(
            reverse("test_login_2"),
            data={"next": f"https://{host}/foo/bar"},
            SERVER_NAME=host,
        )
        self.assertEqual(response.status_code, 302)
        location = response.headers["Location"]
        elements = location.split("?")
        query_string = elements[1]
        arguments = query_string.split("&")
        self.assertIn("client_id=2", arguments)
        self.assertIn(f"redirect_uri=http%3A%2F%2F{host}%2Fcallback-wtf%2F", arguments)



import logging

from django.contrib import messages
from django.http import HttpResponse
from django.template import loader

from django_pyoidc.views import OIDCView

logger = logging.getLogger(__name__)


class OIDCTestSuccessView(OIDCView):
    """
    Used in e2e tests to validate login of a user.

    """

    http_method_names = ["get"]

    def get(self, request, *args, **kwargs):
        messages.success(request, f"message: {request.user.email} is logged in.")

        op_name = self.get_setting("op_name")
        if op_name[:5] == "lemon":
            number = op_name[-1]
            context = {
                "op_login_url": f"e2e_test_ll_login_{number}",
                "op_logout_url": f"e2e_test_ll_logout_{number}",
            }
        else:
            number = op_name[-1]
            context = {
                "op_login_url": f"e2e_test_login_{number}",
                "op_logout_url": f"e2e_test_logout_{number}",
            }
        template = loader.get_template("tests.html")
        return HttpResponse(template.render(request=request, context=context))


class OIDCTestFailureView(OIDCView):
    """
    Used in e2e tests to validate login failures.

    """

    http_method_names = ["get"]

    def get(self, request, *args, **kwargs):
        messages.error(request, "message: something went bad.")

        op_name = self.get_setting("op_name")
        if op_name[:5] == "lemon":
            number = op_name[-1]
            context = {
                "op_login_url": f"e2e_test_ll_login_{number}",
                "op_logout_url": f"e2e_test_ll_logout_{number}",
            }
        else:
            number = op_name[-1]
            context = {
                "op_login_url": f"e2e_test_login_{number}",
                "op_logout_url": f"e2e_test_logout_{number}",
            }

        template = loader.get_template("tests.html")
        return HttpResponse(template.render(request=request, context=context))


class OIDCTestLogoutView(OIDCView):
    """
    Used in e2e tests to validate logout of a user.

    """

    http_method_names = ["get"]

    def get(self, request, *args, **kwargs):
        messages.success(request, "message: post logout view.")

        op_name = self.get_setting("op_name")
        if op_name[:5] == "lemon":
            number = op_name[-1]
            context = {
                "op_login_url": f"e2e_test_ll_login_{number}",
                "op_logout_url": f"e2e_test_ll_logout_{number}",
            }
        else:
            number = op_name[-1]
            context = {
                "op_login_url": f"e2e_test_login_{number}",
                "op_logout_url": f"e2e_test_logout_{number}",
            }
        template = loader.get_template("tests.html")
        return HttpResponse(template.render(request=request, context=context))



import http.client as http_client
import logging
import time
from urllib.parse import parse_qs, urlparse

import requests
from django.urls import reverse
from selenium.webdriver import FirefoxProfile
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.firefox.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from tests.e2e.utils import OIDCE2EKeycloakTestCase

logger = logging.getLogger(__name__)


# HTTP debug for requests
http_client.HTTPConnection.debuglevel = 1


class KeycloakTestCase(OIDCE2EKeycloakTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        options = Options()
        options.add_argument("--private")
        # options.headless = True
        profile = FirefoxProfile()
        profile.set_preference("browser.privatebrowsing.autostart", True)
        options.profile = profile
        cls.selenium = WebDriver(options=options)
        # cls.selenium.implicitly_wait(10)

    @classmethod
    def tearDownClass(cls):
        cls.selenium.quit()
        super().tearDownClass()

    def test_001_m2m_client_credential_success(self, *args):
        """
        Check that we can request the API using a 'service account'.

        We use a 'service account', so we are an external M2M client
        in this situation, we do not use the Django OIDC credentials
        but the creds of a service account, as would be done by an external
        application in a B2B call.
        """

        sso_url = (
            "http://localhost:8080/auth/realms/realm1/protocol/openid-connect/token"
        )
        params = {
            "client_id": "app_m2m1",
            "client_secret": "secret_app-m2m1",
            "grant_type": "client_credentials",
        }
        print("sending M2M Login Request")
        response = requests.post(sso_url, data=params)
        # failing the test in case of bad HTTP status
        response.raise_for_status()

        print("Auth success")
        data = response.json()
        access_token = data["access_token"]

        api_url = f"{self.live_server_url}/api/users"
        headers = {
            "Authorization": f"Bearer {access_token}",
        }
        print("sending API Request with access token")
        response = requests.get(api_url, headers=headers)
        response.raise_for_status()
        content = response.text
        self.assertIn('"username":"service-account-app_m2m1"', content)

    def test_002_m2m_client_credential_forbidden(self, *args):
        """
        Check that we can request the API using a 'service account' and be rejected.

        The client used here does not have access to the same scope as the client
        used in the API. THe group and roles are OK, but not the scope. So Keycloak
        will not add the right audience in the token and we should be rejected.
        """

        sso_url = (
            "http://localhost:8080/auth/realms/realm1/protocol/openid-connect/token"
        )
        params = {
            "client_id": "app2_m2m2",
            "client_secret": "secret_app2-m2m2",
            "grant_type": "client_credentials",
        }
        print("sending M2M Login Request")
        response = requests.post(sso_url, data=params)
        # failing the test in case of bad HTTP status
        response.raise_for_status()

        print("Auth success")
        data = response.json()
        access_token = data["access_token"]

        api_url = f"{self.live_server_url}/api/users"
        headers = {
            "Authorization": f"Bearer {access_token}",
        }
        print("sending API Request with access token, should get a 403 Forbidden.")
        response = requests.get(api_url, headers=headers)
        self.assertEqual(403, response.status_code)

    def test_003_m2m_anonymous_api_access(self, *args):

        # auth part is forbidden
        api_url = f"{self.live_server_url}/api/users"
        response = requests.get(api_url)
        self.assertEqual(403, response.status_code)

        # public part is OK
        api_url = f"{self.live_server_url}/api/publics"
        print("sending anonymous API Request.")
        response = requests.get(api_url)
        response.raise_for_status()
        content = response.text
        self.assertIn("[]", content)

    def test_100_login_page_redirects_to_keycloak_sso(self, *args):
        """
        Test that accessing login callback redirects to the SSO server.
        """
        self.selenium.delete_all_cookies()
        logging.basicConfig()
        logging.getLogger().setLevel(logging.DEBUG)
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True

        client = requests

        print(f"Running Django on {self.live_server_url}")

        login_url = reverse("e2e_test_login_1")
        response = client.get(
            f"{self.live_server_url}{login_url}", allow_redirects=False
        )
        # we should get something like http://localhost:8080/auth/realms/realm1/protocol/openid-connect/auth
        #   ?client_id=app1
        #   &nonce=YZA...24
        #   &redirect_uri=http%3A%2F%2Ftestserver%2Fcallback
        #   &response_type=code
        #   &scope=openid
        #   &state=fe874.....538e047a0
        self.assertEqual(response.status_code, 302)
        location = response.headers["Location"]
        parsed = urlparse(location)
        qs = parse_qs(parsed.query)
        self.assertEqual(parsed.scheme, "http")
        self.assertEqual(parsed.netloc, "localhost:8080")
        self.assertEqual(
            parsed.path, "/auth/realms/realm1/protocol/openid-connect/auth"
        )
        self.assertEqual(qs["client_id"][0], "app1")
        self.assertEqual(qs["redirect_uri"][0], f"{self.live_server_url}/callback-1/")
        self.assertEqual(qs["response_type"][0], "code")
        self.assertEqual(qs["scope"][0], "openid")
        self.assertTrue(qs["state"][0])
        self.assertTrue(qs["nonce"][0])

    def _selenium_front_sso_login(self, user, password):
        front_url = "http://localhost:9999"
        self.selenium.get(front_url)
        WebDriverWait(self.selenium, 30).until(
            EC.element_to_be_clickable((By.ID, "loginBtn"))
        ).click()
        # self.selenium.find_element(By.ID, "loginBtn").click()
        self.wait.until(EC.url_changes(front_url))
        username_input = self.selenium.find_element(By.NAME, "username")
        username_input.send_keys(user)
        password_input = self.selenium.find_element(By.NAME, "password")
        password_input.send_keys(password)
        self.selenium.find_element(By.ID, "kc-login").click()
        self.wait.until(EC.url_matches(front_url))

    def _selenium_front_logout(self):
        front_url = "http://localhost:9999"
        WebDriverWait(self.selenium, 30).until(
            EC.element_to_be_clickable((By.ID, "logoutBtn"))
        ).click()
        self.wait.until(EC.url_matches(front_url))
        bodyText = self.selenium.find_element(By.ID, "message").text
        # check we are NOT logged in
        self.assertEqual("", bodyText)

    def _selenium_sso_login(
        self,
        login_start_url,
        login_end_url,
        user,
        password,
        active_sso_session=False,
        wait_for_success=True,
    ):
        if not active_sso_session:
            self.selenium.get(login_start_url)
            self.wait.until(EC.url_changes(login_start_url))
            # wait ->
            #   EC.url_changes('a') // anything but a
            #   EC.url_contains('')
            #   EC.url_matches(pattern)
            #   EC.title_contains('')
            #   EC.title_is('')
            #   EC.url_to_be('') // exact

            # Uncomment if you want time to detect why it's not working
            WebDriverWait(self.selenium, 30).until(
                EC.presence_of_element_located((By.NAME, "username"))
            )

            username_input = self.selenium.find_element(
                By.NAME,
                "username",
            )
            username_input.send_keys(user)
            password_input = self.selenium.find_element(By.NAME, "password")
            password_input.send_keys(password)
            self.selenium.find_element(By.ID, "kc-login").click()
        else:
            # current SSO session is still active,
            # so we should be redirected directly to success page
            self.selenium.get(login_start_url)
        if wait_for_success:
            self.wait.until(EC.url_matches(login_end_url))

    def _selenium_anon_logout(self):
        """Like classical logout but we are not connected in local app."""
        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we have the logout link
        self.assertTrue("OIDC-ANON-LOGOUT-LINK" in bodyText)
        # click logout
        self.selenium.find_element(By.ID, "oidc-anon-logout-link").click()
        # FIXME: ensure the anon logout worked for real
        # WebDriverWait(self.selenium, 30).until(
        #     EC.presence_of_element_located((By.NAME, "username"))
        # )
        # self.wait.until(EC.url_matches(logout_end_url))

    def _selenium_logout(self, end_url):
        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we have the logout link
        self.assertTrue("OIDC-LOGOUT-LINK" in bodyText)
        # click logout
        self.selenium.find_element(By.ID, "oidc-logout-link").click()
        self.wait.until(EC.url_matches(end_url))
        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we are NOT logged in
        self.assertTrue("You are logged out" in bodyText)

    def test_101_selenium_sso_login(self, *args):
        """
        Test a complete working OIDC login/logout.
        """
        self.selenium.delete_all_cookies()
        timeout = 5
        login_url = reverse("e2e_test_login_1")
        success_url = reverse("e2e_test_success_1")
        post_logout_url = reverse("e2e_test_logout_done_1")
        start_url = f"{self.live_server_url}{login_url}"
        middle_url = f"{self.live_server_url}{success_url}"
        end_url = f"{self.live_server_url}{post_logout_url}"
        self.wait = WebDriverWait(self.selenium, timeout)
        self._selenium_sso_login(
            start_url, middle_url, "user1", "passwd1", active_sso_session=False
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we are logged in
        self.assertTrue("You are logged in as user1@example.com" in bodyText)
        self.assertFalse("You are logged out" in bodyText)

        # Check the session message is shown
        self.assertTrue("message: user1@example.com is logged in." in bodyText)

        # check we have the logout link
        self.assertFalse("OIDC-LOGIN-LINK" in bodyText)
        self.assertTrue("OIDC-LOGOUT-LINK" in bodyText)

        # click logout
        self.selenium.find_element(By.ID, "oidc-logout-link").click()

        self.wait.until(EC.url_matches(end_url))

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text

        # check we are NOT logged in
        self.assertFalse("You are logged in as user1@example.com" in bodyText)
        self.assertTrue("You are logged out" in bodyText)

        # Check the logout view message is shown
        self.assertTrue("message: post logout view." in bodyText)
        # check we have the login link
        self.assertTrue("OIDC-LOGIN-LINK" in bodyText)
        self.assertFalse("OIDC-LOGOUT-LINK" in bodyText)

    def _test_102_selenium_sso_login_relogin_and_logout_with_mixed_clients_id(
        self, *args
    ):
        """
        Test using sso1 and sso2 with different client_id on same sso server.
        """
        self.selenium.delete_all_cookies()
        timeout = 5
        login_url = reverse("e2e_test_login_2")
        success_url = reverse("e2e_test_success_2")
        start_url = f"{self.live_server_url}{login_url}"
        middle_url = f"{self.live_server_url}{success_url}"
        self.wait = WebDriverWait(self.selenium, timeout)

        # LOGIN 1 on sso2
        self._selenium_sso_login(
            start_url, middle_url, "user1", "passwd1", active_sso_session=False
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we are logged in
        self.assertTrue("You are logged in as user1@example.com" in bodyText)
        self.assertFalse("You are logged out" in bodyText)

        # Check the session message is shown
        self.assertTrue("message: user1@example.com is logged in." in bodyText)

        # LOGIN 2: reusing existing session on sso2
        self._selenium_sso_login(start_url, middle_url, "", "", active_sso_session=True)

        # check we are logged in
        self.assertTrue("You are logged in as user1@example.com" in bodyText)
        self.assertFalse("You are logged out" in bodyText)

        # LOGIN 3: using sso1, different client_id on same sso
        # we should get automatcally connected without a form submission
        login_url1 = reverse("e2e_test_login_1")
        success_url1 = reverse("e2e_test_success_1")
        post_logout_url1 = reverse("e2e_test_logout_done_1")
        start_url1 = f"{self.live_server_url}{login_url1}"
        middle_url1 = f"{self.live_server_url}{success_url1}"
        end_url1 = f"{self.live_server_url}{post_logout_url1}"
        self._selenium_sso_login(
            start_url1, middle_url1, "", "", active_sso_session=True
        )

        # check we are logged in
        self.assertTrue("You are logged in as user1@example.com" in bodyText)
        self.assertFalse("You are logged out" in bodyText)

        # check we have the logout link
        self.assertFalse("OIDC-LOGIN-LINK" in bodyText)
        self.assertTrue("OIDC-LOGOUT-LINK" in bodyText)

        # click logout
        self.selenium.find_element(By.ID, "oidc-logout-link").click()

        self.wait.until(EC.url_matches(end_url1))

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text

        # check we are NOT logged in
        self.assertFalse("You are logged in as user1@example.com" in bodyText)
        self.assertTrue("You are logged out" in bodyText)

        # Check the logout view message is shown
        self.assertTrue("message: post logout view." in bodyText)
        # check we have the login link
        self.assertTrue("OIDC-LOGIN-LINK" in bodyText)
        self.assertFalse("OIDC-LOGOUT-LINK" in bodyText)

    def _test_103_selenium_sso_login__relogin_and_logout(self, *args):
        """
        Test a login/logout session, adding a re-login on existing session in the middle
        """
        self.selenium.delete_all_cookies()
        timeout = 5
        login_url = reverse("e2e_test_login_1")
        success_url = reverse("e2e_test_success_1")
        post_logout_url = reverse("e2e_test_logout_done_1")
        start_url = f"{self.live_server_url}{login_url}"
        middle_url = f"{self.live_server_url}{success_url}"
        end_url = f"{self.live_server_url}{post_logout_url}"
        self.wait = WebDriverWait(self.selenium, timeout)

        # LOGIN 1
        self._selenium_sso_login(
            start_url, middle_url, "user1", "passwd1", active_sso_session=False
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we are logged in
        self.assertTrue("You are logged in as user1@example.com" in bodyText)
        self.assertFalse("You are logged out" in bodyText)

        # Check the session message is shown
        self.assertTrue("message: user1@example.com is logged in." in bodyText)

        # LOGIN 2: reusing existing session
        self._selenium_sso_login(start_url, middle_url, "", "", active_sso_session=True)

        # check we are logged in
        self.assertTrue("You are logged in as user1@example.com" in bodyText)
        self.assertFalse("You are logged out" in bodyText)

        # check we have the logout link
        self.assertFalse("OIDC-LOGIN-LINK" in bodyText)
        self.assertTrue("OIDC-LOGOUT-LINK" in bodyText)

        # click logout
        self.selenium.find_element(By.ID, "oidc-logout-link").click()

        self.wait.until(EC.url_matches(end_url))

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text

        # check we are NOT logged in
        self.assertFalse("You are logged in as user1@example.com" in bodyText)
        self.assertTrue("You are logged out" in bodyText)

        # Check the logout view message is shown
        self.assertTrue("message: post logout view." in bodyText)
        # check we have the login link
        self.assertTrue("OIDC-LOGIN-LINK" in bodyText)
        self.assertFalse("OIDC-LOGOUT-LINK" in bodyText)

    def _test_104_selenium_sso_session_with_callbacks(self, *args):
        self.selenium.delete_all_cookies()
        timeout = 5
        login_url = reverse("e2e_test_login_1")
        success_url = reverse("e2e_test_success_1")
        post_logout_url = reverse("e2e_test_logout_done_1")
        start_url = f"{self.live_server_url}{login_url}"
        middle_url = f"{self.live_server_url}{success_url}"
        end_url = f"{self.live_server_url}{post_logout_url}"
        self.wait = WebDriverWait(self.selenium, timeout)
        # previous test destroyed the SSO session, we need to reconnect
        self._selenium_sso_login(
            start_url, middle_url, "user1", "passwd1", active_sso_session=False
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # print(bodyText)

        # Check the session message is shown
        self.assertTrue("message: user1@example.com is logged in." in bodyText)
        # Check the CALLBACK session message is also shown
        self.assertTrue("login_callback for : user1@example.com" in bodyText)

        # check we have the logout link
        self.assertTrue("OIDC-LOGOUT-LINK" in bodyText)
        # click logout
        self.selenium.find_element(By.ID, "oidc-logout-link").click()

        self.wait.until(EC.url_matches(end_url))

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check we are NOT logged in
        self.assertTrue("You are logged out" in bodyText)

        # Check the new session message is shown
        self.assertTrue("message: post logout view." in bodyText)
        # Check the logout callback message is shown
        self.assertTrue("Logout Callback for user1@example.com." in bodyText)

    def _test_105_selenium_sso_failed_login(self, *args):
        """
        Test a failed SSO login (bad client_id: bad_client_id)
        """
        self.selenium.delete_all_cookies()
        timeout = 30
        login_url = reverse("e2e_test_login_3")
        start_url = f"{self.live_server_url}{login_url}"
        wait = WebDriverWait(self.selenium, timeout)
        # wait ->
        #   EC.url_changes('a') // anything but a
        #   EC.url_contains('')
        #   EC.url_matches(pattern)
        #   EC.title_contains('')
        #   EC.title_is('')
        #   EC.url_to_be('') // exact

        self.selenium.get(start_url)
        # wait for the SSO redirection
        wait.until(EC.url_changes(start_url))

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        # check the SSO rejected our client id
        logger.error("*****************************")
        logger.error(bodyText)
        self.assertTrue("We are sorry..." in bodyText)

    def test_106_selenium_ressource_access_checks(self, *args):
        """
        Check that a resource access check can be performed to prevent access for some users.

        @see tests.e2e.test_app.callback:get_user_with_resource_access_check
        """
        self.selenium.delete_all_cookies()
        timeout = 5
        login_location = reverse("e2e_test_login_4")
        success_location = reverse("e2e_test_success_4")
        failure_location = reverse("e2e_test_failure_4")
        post_logout_location = reverse("e2e_test_logout_done_4")
        start_url = f"{self.live_server_url}{login_location}"
        ok_url = f"{self.live_server_url}{success_location}"
        failure_url = f"{self.live_server_url}{failure_location}"
        end_url = f"{self.live_server_url}{post_logout_location}"
        self.wait = WebDriverWait(self.selenium, timeout)

        # user1 login is OK, resource access success
        self._selenium_sso_login(
            start_url,
            ok_url,
            "user1",
            "passwd1",
            active_sso_session=False,
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text

        # Check the session message is shown
        self.assertTrue("message: user1@example.com is logged in." in bodyText)
        self._selenium_logout(end_url)

        # user_limit_app1 has access to app1 and app1-api, resource access should match
        self._selenium_sso_login(
            start_url,
            ok_url,
            "user_limit_app1",
            "passwd1",
            active_sso_session=False,
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        self.assertTrue(
            "message: user_limit_app1@example.com is logged in." in bodyText
        )

        self._selenium_logout(end_url)

        # user_limit_app2 login Will FAIL, as user has access to group2 apps only, and we are not in
        self._selenium_sso_login(
            start_url,
            failure_url,
            "user_limit_app2",
            "passwd2",
            active_sso_session=False,
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        print(bodyText)
        # Check we have the Permission Denied message
        self.assertTrue("Permission Denied." in bodyText)
        # we are not allowed in this app but we still have a valid SSO session
        # so having a logout action can be useful
        self._selenium_anon_logout()

        # to ensure this destroy cookies worked redo a user1 login/logout
        # user1 login is OK, resource access success
        self._selenium_sso_login(
            start_url,
            ok_url,
            "user1",
            "passwd1",
            active_sso_session=False,
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text

        # Check the session message is shown
        self.assertTrue("message: user1@example.com is logged in." in bodyText)
        self._selenium_logout(end_url)

    def test_107_selenium_minimal_audience_checks(self, *args):
        """
        Check that a minimal audience check can be performed to prevent access for some users.
        @see tests.e2e.test_app.callback:get_user_with_minimal_audiences_check
        """
        self.selenium.delete_all_cookies()
        timeout = 5
        login_location = reverse("e2e_test_login_5")
        success_location = reverse("e2e_test_success_5")
        post_logout_location = reverse("e2e_test_logout_done_5")
        start_url = f"{self.live_server_url}{login_location}"
        ok_url = f"{self.live_server_url}{success_location}"
        end_url = f"{self.live_server_url}{post_logout_location}"
        self.wait = WebDriverWait(self.selenium, timeout)

        # user1 login is OK, resource access success
        self._selenium_sso_login(
            start_url,
            ok_url,
            "user1",
            "passwd1",
            active_sso_session=False,
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text

        # Check the session message is shown
        self.assertTrue("message: user1@example.com is logged in." in bodyText)
        self._selenium_logout(end_url)

        # user_limit_app1 has access to app1 and app1-api, aud will be app1-api
        # should be OK here
        self._selenium_sso_login(
            start_url,
            ok_url,
            "user_limit_app1",
            "passwd1",
            active_sso_session=False,
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text
        self.assertTrue(
            "message: user_limit_app1@example.com is logged in." in bodyText
        )

        self._selenium_logout(end_url)

        # WARRN: FAILURE to detect bad user here, but that's because the HOOK_USER used
        # is not good enough for Keycloak, a ressource_access check would be better.
        # user_limit_app2 login Will NOT FAIL
        # as our 'aud' check is not sufficient to detect that, sad...
        # since keycloak decided to remove the current client_di in 'aud' in access_tokens
        # that's hard to detect this type of users from access_tokens without test resource_access
        self._selenium_sso_login(
            start_url,
            ok_url,
            "user_limit_app2",
            "passwd2",
            active_sso_session=False,
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text

        # Check the session message is shown
        self.assertTrue(
            "message: user_limit_app2@example.com is logged in." in bodyText
        )
        self._selenium_logout(end_url)

        # And then test the user with only one app access
        # to ensure that the 'aud' token is still present for him even if the audience is just
        # the present app.
        self._selenium_sso_login(
            start_url,
            ok_url,
            "user_app1_only",
            "passwd1",
            active_sso_session=False,
        )

        bodyText = self.selenium.find_element(By.TAG_NAME, "body").text

        # Check the session message is shown
        self.assertTrue("message: user_app1_only@example.com is logged in." in bodyText)
        self._selenium_logout(end_url)

    def test_201_selenium_front_app_api_call(self, *args):
        """
        Check that a test front app can make an OIDC API call.
        """
        timeout = 60
        self.wait = WebDriverWait(self.selenium, timeout)

        # user1 login is OK, resource access success
        self._selenium_front_sso_login("user1", "passwd1")
        # let the page reloads after login fill the user session stuff
        time.sleep(3)
        bodyText = self.selenium.find_element(By.ID, "message").get_attribute(
            "innerHTML"
        )
        self.assertTrue("User: user1" in bodyText)

        WebDriverWait(self.selenium, 30).until(
            EC.element_to_be_clickable((By.ID, "securedBtn"))
        ).click()

        # let the ajax stuff behave
        time.sleep(3)
        bodyText = self.selenium.find_element(By.ID, "message").get_attribute(
            "innerHTML"
        )
        logger.error(bodyText)
        self.assertTrue("user1@example.com" in bodyText)

        # FIXME: there a selenium issue in the logout btn selection...
        self._selenium_front_logout()

        # After logout, launch unauthorized ajax call
        WebDriverWait(self.selenium, 30).until(
            EC.element_to_be_clickable((By.ID, "securedBtn"))
        ).click()
        # let the ajax stuff behave
        time.sleep(3)
        bodyText = self.selenium.find_element(By.ID, "message").get_attribute(
            "innerHTML"
        )
        self.assertTrue("Request Forbidden" in bodyText)
